/**
 * $Id: minheap.c 22 2006-05-03 20:12:20Z alexeen $
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "minheap.h"

/** Private functions. **/
static void minheap_rebuild(minheap_t *h);
static void minheap_heapify(minheap_t *h, int i);
static void minheap_swap(minheap_t *h, int s1, int s2);
static void minheap_remove_min(minheap_t *h);

/**
 * minheap_init: Initializes a new min-heap data
 *   structure.
 * @return Returns a malloc'ed min-heap data
 *   structure, or NULL pointer if failed.
 */
minheap_t *minheap_init(void)
{
	minheap_t *h;

	h = (minheap_t *)malloc(sizeof(minheap_t));
	if (!h) { return NULL; }

	/** Initial values. **/
	h->size = 0;
	h->s_size = 0;
	h->heap = NULL;

	return h;
}

/**
 * minheap_insert: Inserts a new min-heap element
 *   into the heap.
 * h - The min-heap to insert into.
 * key - The key uniquely identify this piece of data.
 * data - The data to store into this min-heap.
 * @return Returns 0 for success, and -1 otherwise.
 */
int minheap_insert(minheap_t *h, void *key, void *data)
{
	int inc = 0;
	minheap_elm_t *new = NULL;
	minheap_elm_t *dst = NULL;

	if (!h || !data || h->size > h->s_size) { return -1; }

	/** Increase heap size. **/
	if (h->size == h->s_size)
	{
		/** Double the current size every time. **/
		inc = h->s_size < 1 ? 1 : h->s_size*2;

		new = (minheap_elm_t *)realloc(h->heap, sizeof(minheap_elm_t)*inc);
		if (!new) { return -1; }

		h->heap = new; /** Use the new pointer. **/
		h->s_size = inc;
	}

	/** Insert the new data into min-heap. **/
	dst = h->heap + h->size;
	dst->key = key;
	dst->data = data;
	h->size++;

	/** Fix up the heap structure. **/
	minheap_rebuild(h);

	return 0;
}

/**
 * minheap_extract_min: Extracts the minimum min-heap
 *   element from the min-heap.
 * h - The min-heap to extract from.
 * @return Returns a pointer to the data of the minimum
 *   min-heap element of the min-heap `h', or
 *   NULL pointer if errored.
 */
void *minheap_extract_min(minheap_t *h)
{
	void *data = NULL;
	minheap_elm_t *heap = NULL;

	if (!h || h->size < 1) { return data; }
	heap = h->heap;
	data = heap->data;

	minheap_remove_min(h);

	return data;
}

/**
 * minheap_destroy: Frees all allocated resources for
 *   the specified min-heap. The data pointed to by
 *   the data pointer will not be freed by this
 *   function call. If user does not free the data
 *   manually before hand or have other means of
 *   accessing those resources, this function call
 *   may result in possible memory leaks.
 * h - The min-heap to destroy.
 */
void minheap_destroy(minheap_t *h)
{
	if (h && h->heap)
	{
		free(h->heap);
		h->heap = NULL;
		h->s_size = 0;
		h->size = 0;
		free(h);
	}
}

/**
 * minheap_size: Returns the tree size of
 *   the min-heap.
 * h - The min-heap to look at.
 * @return Returns the tree size of the
 *   min-heap.
 */
size_t minheap_size(minheap_t *h)
{
	if (!h) { return -1; }
	return h->size;
}

/**
 * minheap_s_size: Returns the total available
 *   storage size allocated for this min-heap.
 * h - The min-heap to look at.
 * @return Returns the storage size of the
 *   min-heap.
 */
size_t minheap_s_size(minheap_t *h)
{
	if (!h) { return -1; }
	return h->s_size;
}

/**
 * minheap_remove_min: Removes the minimum element
 *   from the min-heap tree.
 * h - The min-heap to remove from.
 */
static void minheap_remove_min(minheap_t *h)
{
	int end_ind = h->size-1;

	minheap_elm_t *heap = h->heap;
	minheap_elm_t *last = heap+end_ind;

	free(heap->key); /** Avoid memory leaks. **/
	heap->key = last->key;
	heap->data = last->data;

	/**
	 * Instead of shrinking the allocated resources,
	 * simply mark it as unused to avoid frequent calls
	 * of realloc.
	 */
	last->key = NULL;
	last->data = NULL;
	h->size--;

	/** Maintains consistency of min-heap. **/
	minheap_rebuild(h);
}

/**
 * minheap_rebuild: Restructure the heap wherever necessary
 *   to maintain the consistency.
 * h - Heap to heapify.
 */
static void minheap_rebuild(minheap_t *h)
{
	/** Location in min-heap. **/
	int ind = (int) (floor((double) h->size)/2.0);
	int i;

	/** Run heapify on all non-leaf nodes. **/
	for (i = ind; i > 0; i--)
	{
		minheap_heapify(h, i);
	}
}

/**
 * minheap_heapify: Perculates the element at
 *   index i downwards until the subtree of this
 *   element being stored at becomes a proper min-heap.
 * h - Heap to heapify on.
 * i - The index of the heap element to heapify on.
 */
static void minheap_heapify(minheap_t *h, int i)
{
	int smallest = i;

	minheap_elm_t *heap = h->heap;

	minheap_elm_t *tmp1 = heap + ((2*i)-1);
	minheap_elm_t *tmp2 = heap + (i-1);

	/** Checks the left side. **/
	if (2*i <= h->size && 0 > minheap_compare(tmp1->key, tmp2->key))
	{
		/** Left side is perfered so far. **/
		smallest = 2*i;
	}

	tmp1 = heap + (2*i);
	tmp2 = heap + (smallest-1);

	/**
	 * Checks the right side as well to see which side
	 * to perculate down to.
	 **/
	if ((2*i)+1 <= h->size && 0 > minheap_compare(tmp1->key, tmp2->key))
	{
		/** Right side is perfered over the left. **/
		smallest = (2*i)+1;
	}

	/** Make sure there's a need to perculate downwards. **/
	if (smallest != i)
	{
		/** Swap the positions of the elements. **/
		minheap_swap(h, i-1, smallest-1);

		/** Recursively go down the subtree. **/
		minheap_heapify(h, smallest);
	}
}

/**
 * minheap_swap: Swaps two element in the storage array.
 * h - The min-heap to work with.
 * s1 - The index of the element to swap with s2.
 * s2 - The index of the element to swap with s1.
 */
static void minheap_swap(minheap_t *h, int s1, int s2)
{
	void *key = NULL;
	void *data = NULL;

	minheap_elm_t *e1 = h->heap + s1;
	minheap_elm_t *e2 = h->heap + s2;

	key = e1->key;
	data = e1->data;

	e1->key = e2->key;
	e1->data = e2->data;

	e2->key = key;
	e2->data = data;
}

/**
 * minheap_print: Outputs the current state of the
 *   min-heap tree.
 * h - The min-heap to display.
 */
void minheap_print(minheap_t *h)
{
	int i = 0;
	minheap_elm_t *heap = NULL;

	if (!h) { printf("[ Invalid ]\n"); }
	else
	{
		heap = h->heap;

		printf("( ");

		for (i = 0; i < h->size; i++)
		{
			printf("[ %s ] ", (char *) heap->data);
			heap++;
		}

		printf(")\n");
	}
}
